VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "CPython"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Option Explicit

'Author:  David Zimmer <dzzie@yahoo.com>
'Site:    http://sandsprite.com


Implements IPythonScript  'same interface for callbacks for class or debugger ui

Event devmsg(m As String)
Event ScriptOut(msg As String)
Event ScriptError(msg As String)
Event TimeOut(ByRef Cancel As Boolean)
Event DebugHandler(ByVal t As dcb_type, msg As String)
Event CmdOutput(m As String)

Public api As New CPyApi
Public pythonReady As Boolean

Private mSharedObjects As New Collection
Private mIsRunning As Boolean
Private set_trace As String   'for running with debugger...
Private pyComImported As Boolean

'one of these two adds to the class ref count
Private WithEvents tmrTimeout As Timer
Attribute tmrTimeout.VB_VarHelpID = -1
Private mTmrHost As frmTimeOutHost
Private mInitilized As Boolean
Private mPythonPath As String
Private WithEvents cmd As CCmdOutput

Property Get homeDir() As String
    homeDir = mPythonPath
End Property

Function pyCmd(ByVal cmdLine As String) As String
    
    Dim exe As String
    
    exe = mPythonPath & "\python.exe"
    If Not FileExists(exe) Then Exit Function
    
    If cmd.GetCommandOutput(exe, cmdLine) Then
        pyCmd = cmd.result
    End If
    
End Function

'example: pipCmd("install regex")
Function pipCmd(ByVal cmdLine As String) As String
    
    Dim exe As String
    
    exe = mPythonPath & "\Scripts\pip.exe"
    If Not FileExists(exe) Then Exit Function
    
    If cmd.GetCommandOutput(exe, cmdLine) Then
        pipCmd = cmd.result
    End If
    
End Function

Private Sub Class_Initialize()
    
    Set cmd = New CCmdOutput
    
    set_trace = "sys.settrace(None)" & vbCrLf & _
                "import inspect" & vbCrLf & _
                "curFrame = None" & vbCrLf & _
                "def my_tracer(frame, event, arg = None):" & vbCrLf & _
                "    global curFrame" & vbCrLf & _
                "    curFrame = frame" & vbCrLf & _
                "    code = frame.f_code" & vbCrLf & _
                "    line_no = frame.f_lineno" & vbCrLf & _
                "    filename = code.co_filename #will be <string> for our SimpleString evals" & vbCrLf & _
                "    func_name = code.co_name" & vbCrLf & _
                "    caller = frame.f_back" & vbCrLf & _
                "    if not caller is None:" & vbCrLf & _
                "        caller_line_no = caller.f_lineno" & vbCrLf & _
                "        caller_func_name = caller.f_code.co_name" & vbCrLf & _
                "        caller_filename = caller.f_code.co_filename" & vbCrLf & _
                "    else:" & vbCrLf & _
                "        caller_line_no = 0;caller_func_name='';caller_filename=''" & vbCrLf & _
                "  " & vbCrLf

     set_trace = set_trace & _
                "    if str(line_no) == '0': return my_tracer" & vbCrLf & _
                "    if str(func_name) == ""flush"": return my_tracer" & vbCrLf & _
                "    if str(func_name) == ""write"": return my_tracer" & vbCrLf & _
                "    if str(filename) != ""<string>"": return my_tracer" & vbCrLf & _
                "  " & vbCrLf & _
                "    emb.dbgOut(0,f""{line_no},{event},{func_name},{filename},{caller_line_no},{caller_func_name},{caller_filename}"")" & vbCrLf & _
                "    " & vbCrLf & _
                "    return my_tracer" & vbCrLf & _
                "  " & vbCrLf & _
                "isDebugging = 1" & vbCrLf & _
                "sys.settrace(my_tracer) #no more lines after this or it will be traced" & vbCrLf
                
End Sub
 
Property Get Initilized() As Boolean
    Initilized = mInitilized
End Property

Function ShowAbout(Optional display As Boolean = True) As String
    
    ShowAbout = About_py4vb & vbCrLf & vbCrLf & GetDllInfos()
    If display Then MsgBox ShowAbout, vbInformation, "About py4vb"
    
End Function

Private Sub cmd_OutPutAvailable(Data As String)
    Dim tmp As String
    tmp = Trim(UnixToDos(Data))
    If Len(tmp) = 0 Then Exit Sub
    RaiseEvent CmdOutput(tmp)
End Sub

Private Function IPythonScript_DebugHandler(ByVal t As dcb_type, msg As String) As Long
    Dim retCode As Long
    RaiseEvent DebugHandler(t, msg)
    IPythonScript_DebugHandler = retCode 'unused but we could pass it by ref in event though..
End Function

'can they call abort on a timer for timeout? we can still default to calling an interrupt may not work though..
Property Get TimeoutAvailable() As Boolean
    TimeoutAvailable = timeoutEnabled 'module level var eventually not available in ocx build
End Property

Private Sub tmrTimeout_Timer()
    Dim Cancel As Boolean
    RaiseEvent TimeOut(Cancel)
    If Not Cancel Then Exit Sub
    AbortScript
End Sub

'automatically disabled on reset and script end
Function SetTimeOut(Optional secs As Long = 0)

    If tmrTimeout Is Nothing And secs < 1 Then Exit Function 'nothing to do its not set yet
    
    If tmrTimeout Is Nothing Then 'lazy load on first use
        Set mTmrHost = New frmTimeOutHost
        mTmrHost.Visible = False
        Set tmrTimeout = mTmrHost.Timer1
    End If
    
    If secs <= 0 Then
        tmrTimeout.Enabled = False
    Else
        'Debug.Print "timeoutEnabled:" & timeoutEnabled
        tmrTimeout.Enabled = False 'resets interval..
        tmrTimeout.Interval = secs * 1000
        tmrTimeout.Enabled = True
    End If
    
End Function

'has to be called manually, the form will keep it alive
'class terminate would never fire even if set class = nothing
Function ReleaseTimer()
    If Not tmrTimeout Is Nothing Then
        tmrTimeout.Enabled = False
        Set tmrTimeout = Nothing
        Unload mTmrHost
        Set mTmrHost = Nothing
    End If
End Function


Friend Property Get IPythonScript_isRunning() As Boolean
    IPythonScript_isRunning = mIsRunning
End Property

Friend Property Get IPythonScript_SharedObjects() As Collection
    Set IPythonScript_SharedObjects = mSharedObjects
End Property

Friend Sub IPythonScript_vbOut(t As cb_type, msg As String)
    If t = cb_devMsg Then RaiseEvent devmsg(msg)
    If t = cb_output Then RaiseEvent ScriptOut(msg)
    If t = cb_error Then RaiseEvent ScriptError(msg)
End Sub



Function abort()
    py4vb.AbortScript True
End Function
    
'you can run this manually or it will be called on demand on first required use
'we can not call it from class initilize because the event sink isnt set yet (kind of a vb bug)
Function InitPython(Optional ByVal pyDll As String) As Boolean
    
    Dim dll As String, hPy As Long
    
    mPythonPath = Empty
     
    If pythonReady Then
        RaiseEvent devmsg("Python 311 has already been loaded")
        Exit Function
    End If
    
    dll = App.path & "\..\py4vb.dll"
    If Not FileExists(dll) Then dll = App.path & "\py4vb.dll"
    
    If FileExists(dll) Then
        RaiseEvent devmsg("Found: " & dll)
    Else
        RaiseEvent devmsg("Failed to find py4vb.dll")
        Exit Function
    End If
    
    If Len(pyDll) > 0 Then
        If FolderExists(pyDll) Then pyDll = pyDll & "\python311.dll"
        If Not FileExists(pyDll) Then Exit Function
    
        If is32Bit(pyDll) Then
            RaiseEvent devmsg("manually set python311.dll confirmed 32 bit")
            'better preload it since it may not be in path..(will it find the imports? dunno better to be in path)
            'we could also use SetDllDirectory safer?
            hPy = LoadLibrary(pyDll)
            If hPy = 0 Then
                RaiseEvent devmsg("Failed to LoadLibrary(python311.dll)")
                Exit Function
            Else
                RaiseEvent devmsg("Manually preloaded python311.dll ok: " & Hex(hPy))
                mPythonPath = GetParentFolder(pyDll)
                If Right(mPythonPath, 1) = "\" Then mPythonPath = Mid(mPythonPath, 1, Len(mPythonPath) - 1)
            End If
        Else
            RaiseEvent devmsg("python311.dll is NOT 32 bit")
            Exit Function
        End If
    Else
        If Not scanPathFor32bitPythonInstall(pyDll) Then
            RaiseEvent devmsg("Could not find 32bit python311.dll in path? probably wont work...") 'but might..
        Else
            RaiseEvent devmsg("32 bit python found in path: " & pyDll)
            mPythonPath = GetParentFolder(pyDll)
            If Right(mPythonPath, 1) = "\" Then mPythonPath = Mid(mPythonPath, 1, Len(mPythonPath) - 1)
        End If
    End If
    
    If Not LoadPyLib(dll) Then
        RaiseEvent devmsg("Failed to load py4vb lib")
        Exit Function
    Else
        RaiseEvent devmsg("Python loaded and ready")
        pythonReady = True
    End If

    pythonReady = True
    InitPython = True
    
End Function

'can delay 5 seconds while waiting for script terminate if it was running...
Function Reset() As Boolean
    
    Dim n As Date
    
    If Not pythonReady Then If Not InitPython Then Exit Function

    If mIsRunning Then
        AbortScript
        n = Now
        Do While DateDiff("s", Now, n) < 5
            DoEvents
            If Not mIsRunning Then Exit Do
        Loop
        If mIsRunning Then Exit Function
    End If
    
    If Not tmrTimeout Is Nothing Then tmrTimeout.Enabled = False
    Set mSharedObjects = New Collection
    Finalize
    Reset = True
    
End Function

Function AddObject(o As Object, ByVal name As String) As Boolean

    On Error Resume Next
    
    Dim rv As Long, b As Boolean
    
    If Not pyComImported Then
        rv = PyOp(op_ImportMod, , "pycom")
        If rv = 0 Then
            RaiseEvent devmsg("AddObject auto import pycom failed")
            'Exit Function
        Else
            pyComImported = True
        End If
    End If

    mSharedObjects.Add o, name
    
    'do this after the above of course...
    b = AddCode(name & " = pycom.GetObject('" & name & "')")
    If Not b Then RaiseEvent devmsg("Addobject " & name & " variable = " & b)
        
    AddObject = (Err.Number = 0)
    
End Function

Function Initilize(Optional withDebugger As Boolean = False) As Boolean
    Dim rv As Long
    
    If Not pythonReady Then If Not InitPython Then Exit Function

    If Not ActiveObject Is Nothing Then
        If ActiveObject.isRunning Then
            RaiseEvent devmsg("Previous Active Object is still running. Reset First")
            Exit Function
        End If
    End If
    
    pyComImported = False
    AbortScript False
    Set ActiveObject = Me 'global in module used for event bubbling
    Set mSharedObjects = New Collection
    rv = PyOp(op_Init)
    Initilize = (rv = 1)
    mInitilized = (rv = 1)
    RaiseEvent devmsg("Initilize = " & rv)
    
    If withDebugger Then
        rv = PyOp(op_LoadCode, , set_trace)
        RaiseEvent devmsg("set_trace = " & rv)
    End If
    
    On Error Resume Next
    'some libs like ctypes have dlls and expect to start from python home dir
    If FolderExists(mPythonPath) Then
        ChDir mPythonPath
        AddCode "sys.path.append(""" & Replace(mPythonPath, "\", "\\") & """)"
        AddCode "sys.path.append(""" & Replace(mPythonPath & "\DLLs", "\", "\\") & """)"
    End If
    
End Function

Function AddFile(ByVal fPath As String, Optional setHomeDir As Boolean = True) As Boolean
    
    Dim rv As Long
    
    If Not mInitilized Then
        RaiseEvent devmsg("AddFile: Not initilized?")
        Exit Function
    End If
    
    If Not FileExists(fPath) Then
        RaiseEvent devmsg("AddFile: path not found: " & fPath)
        Exit Function
    End If
    
    'would require shared VC runtime (/MD) to add to py4vb (and VC runtime install)
    'If Not runFileEnabled Then
    '    RaiseEvent devmsg("AddFile not enabled custom dll not loaded..")
    '    Exit Function
    'End If
    
    If setHomeDir Then ChDir GetParentFolder(fPath)
    'AddCode "sys.path.append(os.getcwd())" 'this would work but req chdir & import os
    
    AddCode "sys.path.append(""" & Replace(GetParentFolder(fPath), "\", "\\") & """)"
    
    'rv = PyRun_SimpleFilePath(fPath) '-1 on failure, 0 success
    rv = PyOp(op_LoadFile, , fPath)
    If rv = 0 Then AddFile = True
    
End Function

Function AddCode(ByVal pyCode As String) As Boolean
    Dim rv As Long
    
    If Not pythonReady Then If Not InitPython Then Exit Function
    
    mIsRunning = True
    rv = PyOp(op_LoadCode, , pyCode)
    mIsRunning = False
    'If Not tmrTimeout Is Nothing Then tmrTimeout.Enabled = False they might call add code on their own more than once..
    AddCode = (rv = 0)
    
End Function

 
Function Eval(ByVal pyCode As String, Optional retPyObj As Boolean = False) As Variant
    
    Dim hRet As Long
    Dim po As CPythonObj
    
    If Not pythonReady Then If Not InitPython Then Exit Function
    mIsRunning = True
    
    hRet = PyOp(op_eval, , pyCode)
    
    If retPyObj Then
        Set po = New CPythonObj
        po.hObj = hRet
        'po.strongRef = True
        Set Eval = po
    ElseIf hRet <> 0 Then
        Eval = getValue(hRet) 'returns a vb type
        decref hRet
    End If
    
    'If Not tmrTimeout Is Nothing Then tmrTimeout.Enabled = False they might call eval on their own more than once..
    mIsRunning = False
    
End Function

Function Finalize() As Boolean
    Dim rv As Long
    If Not pythonReady Then If Not InitPython Then Exit Function
    'todo: what happens if finalize while running? probably bad
    rv = PyOp(op_Final)
    If Not tmrTimeout Is Nothing Then tmrTimeout.Enabled = False
    Finalize = (rv = 0)
    mInitilized = False
End Function



