Attribute VB_Name = "modDbg"
Option Explicit

Global Const SC_MARK_CIRCLE = 0
Global Const SC_MARK_ARROW = 2
Global Const SC_MARK_BACKGROUND = 22
Global Const curFile As String = "<string>" 'we only debug current file

'http://www.scintilla.org/aprilw/SciLexer.bas

Public breakpoints As New Collection
Global running As Boolean
Global waitingToContinue As Boolean
Global lastEIP As Long

Enum runModes
    rm_stepIn = 2
    rm_stepOut = 3
    rm_stepOver = 4
    rm_runToLine = 5
    rm_run = 6
    rm_stop = 7
End Enum

Public Type Stats
    state As runModes
    filename As String
    curFunc As String
    lineNumber As Long
    eventName As String
    callerFuncName As String
    callerFileName As String
    callerLineNum As Long
    stepToLine As Long
    stepOutFunc As String
    stepOverBreakOnFunc As String
    stepOverCallerFrame As String
End Type

Public status As Stats


Public Function BreakPointExists(filename As String, lineNo As Long, Optional b As CBreakpoint, Optional colIndex As Long) As Boolean

    On Error Resume Next
    colIndex = 1
    For Each b In breakpoints
        If b.lineNo = lineNo And b.filename = filename Then
            BreakPointExists = True
            Exit Function
        End If
        colIndex = colIndex + 1
    Next
    
    colIndex = 0
    
End Function

Public Sub ToggleBreakPoint(filename As String, lineNo As Long, sourceText As String)
  
    If BreakPointExists(filename, lineNo) Then
        RemoveBreakpoint filename, lineNo
    Else
        SetBreakpoint filename, lineNo, sourceText
        Debug.Print "Bp on " & lineNo & " : " & sourceText
    End If

End Sub

Public Function SetBreakpoint(ByVal filename As String, lineNo As Long, ByVal sourceText As String) As Boolean
    Dim b As CBreakpoint
    
    If BreakPointExists(filename, lineNo, b) Then
        b.sourceText = sourceText 'just in case it changed..
        SetBreakpoint = True
        Exit Function
    End If
    
    'If Not isExecutableLine(sourceText) Then 'just covers some basics for convience..
    '    doOutput "Can not set breakpoint here, not an executable line"
    '    Exit Function
    'End If
    
    Set b = New CBreakpoint
    
    With b
        .filename = filename
        .lineNo = lineNo
        .sourceText = sourceText
    End With
    
    If running Then
        b.isSet = True
    End If

    breakpoints.Add b
    Form1.scivb.SetMarker lineNo
    SetBreakpoint = True
    
    
End Function

Public Sub RemoveBreakpoint(filename As String, lineNo As Long)
    Dim b As CBreakpoint
    Dim colIndex As Long
    Dim cur_b As CBreakpoint
    
    If Not BreakPointExists(filename, lineNo, b, colIndex) Then Exit Sub
    
    If running Then
        'we have to compact our duktape bp indexes - technically we should call relist...
        'note we specifically dont check filename its a flat array currently
        For Each cur_b In breakpoints
            If cur_b.index > b.index Then
                cur_b.index = cur_b.index - 1
            End If
        Next
    End If
        
    Form1.scivb.DeleteMarker lineNo
    breakpoints.Remove colIndex
    
End Sub

Sub RemoveAllBreakpoints()
    Dim b As CBreakpoint
    For Each b In breakpoints
        RemoveBreakpoint b.filename, b.lineNo
    Next
End Sub


Public Sub SyncUI()
       
    Dim curline As Long
    
    'If Len(status.fileName) = 0 Or status.fileName <> curFile Then Exit Sub
    
    ClearLastLineMarkers
    
    curline = status.lineNumber - 1
    Form1.scivb.SetMarker curline, 1
    Form1.scivb.SetMarker curline, 3
    lastEIP = curline
    
    Form1.scivb.GotoLine curline
    Form1.scivb.SetFocus
    Form1.Refresh
    
    'Python -> C emb.dbgOut -> VB modDbg.cb_cbgOut -> ParseX function below
    Form1.lvVars.Clear
    PyOp op_LoadCode, , "emb.dbgOut(1,str(inspect.stack()))"
    
    'PyOp op_LoadCode, , "dumpArgs()" 'locals contains all args
    
    If Form1.lvCallStack.ListItems.Count > 3 Then
        PyOp op_LoadCode, , "dumpLocals()" 'todo: if not in function, skip locals (= globals at this scope)
    End If
    
    PyOp op_LoadCode, , "dumpGlobals()"
    
    DoEvents

End Sub
  
Sub ParseCallStack(callStack As String)
     
     Dim tmp() As String, t
     tmp = Split(callStack, "FrameInfo")
     
     With Form1
           .lvCallStack.Clear
            For Each t In tmp
                If Len(t) > 1 Then .lvCallStack.AddItem t
            Next
     End With

'    [
'        FrameInfo(
'                frame=<
'                   frame at 0x0B090AB8,
'                   File '<string>',
'                   line 1,
'                   code <module>
'                >,
'                filename='<string>',
'                lineno=1,
'                function='<module>',
'                code_context=None,
'                index=None,
'                positions=Positions(
'                    lineno=1,
'                    end_lineno=1,
'                    col_offset=23,
'                    end_col_offset = 38
'                )
'            ),
'    ]

End Sub

Sub ParseGlobals(globals As String)
    
'__name__=__main__<ITEM_DIVIDER>__doc__=None<ITEM_DIVIDER>__package__=None<ITEM_DIVIDER>
'__loader__=<class '_frozen_importlib.BuiltinImporter'><ITEM_DIVIDER>__spec__=None
'<ITEM_DIVIDER>__annotations__={}<ITEM_DIVIDER>__builtins__=<module 'builtins' (built-in)>
'<ITEM_DIVIDER>sys=<module 'sys' (built-in)><ITEM_DIVIDER>emb=<module 'emb' (built-in)>
'<ITEM_DIVIDER>CatchOut=<class '__main__.CatchOut'><ITEM_DIVIDER>CatchErr=<class '__main__.CatchErr'>
'<ITEM_DIVIDER>myExit=<function myExit at 0x03CD6618><ITEM_DIVIDER>exit=<function myExit at 0x03CD6618>
'<ITEM_DIVIDER>quit=<function myExit at 0x03CD6618><ITEM_DIVIDER>catchOut=<__main__.CatchOut object at 0x03CE4590>
'<ITEM_DIVIDER>catchErr=<__main__.CatchOut object at 0x03CE44F0><ITEM_DIVIDER>isDebugging=1
'<ITEM_DIVIDER>alert=<function alert at 0x03CE8168><ITEM_DIVIDER>dumpG=<function dumpG at 0x03CE81B8>
'<ITEM_DIVIDER>inspect=<module 'inspect' from 'C:\\Py311\\Lib\\inspect.py'>
'<ITEM_DIVIDER>curFrame=<frame at 0x03AF79E8, file '<string>', line 6, code <module>>
'<ITEM_DIVIDER>my_tracer=<function my_tracer at 0x03CE8208>
'<ITEM_DIVIDER>datetime=<class 'datetime.datetime'><ITEM_DIVIDER>

    'todo: filter noise
    
    Dim tmp() As String, t, t2() As String
    Static filter() As String, skipIt As Boolean
    
    On Error Resume Next

    If AryIsEmpty(filter) Then
        filter = Split("__name__,__doc__,__package__,__loader__,__spec__,__annotations__,__builtins__,sys,emb,CatchOut,CatchErr,myExit,exit,quit,catchOut,catchErr,isDebugging,alert,dumpGlobals,dumpLocals,dumpArgs,inspect,curFrame,my_tracer,datetime", ",")
    End If
    
    tmp = Split(globals, "<ITEM_DIVIDER>")
    For Each t In tmp
        skipIt = False
        t2 = Split(t, "=", 2)
        If Form1.chkShowAllVars.value = 1 Then
            Form1.lvVars.AddItem "global", t2(0), t2(1)
        Else
            If AryContains(t2(0), filter) Then skipIt = True
            If Not skipIt And Form1.chkShowFuncs.value = 0 And VBA.Left(t2(1), 9) = "<function" Then skipIt = True
            If Not skipIt Then Form1.lvVars.AddItem "global", t2(0), t2(1)
        End If
    Next
 
End Sub

Function AryContains(elem As String, ary() As String, Optional caseSensitive As Boolean = True) As Boolean
    Dim x, match As Boolean
    
    If AryIsEmpty(ary) Then Exit Function
    
    For Each x In ary
    
        If caseSensitive Then
            If x = elem Then match = True
        Else
            If LCase(x) = LCase(elem) Then match = True
        End If
        
        If match Then
            AryContains = True
            Exit Function
        End If
        
    Next
    
End Function

Sub ParseLocals(locals As String)
    
    'todo: when not in a function, locals will show all globals... filter me
    On Error Resume Next
    Dim tmp() As String, t, t2() As String

    If Len(locals) = 0 Then Exit Sub

    tmp = Split(locals, "<ITEM_DIVIDER>")
    For Each t In tmp
        t2 = Split(t, "=", 2)
        Form1.lvVars.AddItem "locals", t2(0), t2(1)
    Next

End Sub


Sub ParseArgs(args As String)

    On Error Resume Next
    Dim tmp() As String, t, t2() As String
    
    If Len(args) = 0 Then Exit Sub

    tmp = Split(args, "<ITEM_DIVIDER>")
    For Each t In tmp
        t2 = Split(t, "=", 2)
        Form1.lvVars.AddItem "args", t2(0), t2(1)
    Next

End Sub


Public Sub ClearLastLineMarkers()
    Dim startPos As Long, endPos As Long

    Form1.scivb.DeleteMarker lastEIP, 1 'remove the yellow arrow
    Form1.scivb.DeleteMarker lastEIP, 3 'remove the yellow line backcolor

    'force a refresh of the specified line or it might not catch it..
    startPos = Form1.scivb.PositionFromLine(lastEIP)
    endPos = Form1.scivb.PositionFromLine(lastEIP + 1)
    Form1.scivb.DirectSCI.Colourise startPos, endPos
    
End Sub

Function AryIsEmpty(ary) As Boolean
  On Error GoTo oops
  Dim i
    i = UBound(ary)  '<- throws error if not initalized
    AryIsEmpty = False
  Exit Function
oops: AryIsEmpty = True
End Function


Sub push(ary, value) 'this modifies parent ary object
Dim x
    On Error GoTo init
    x = UBound(ary) '<-throws Error If Not initalized
    ReDim Preserve ary(UBound(ary) + 1)
    ary(UBound(ary)) = value
    Exit Sub
init:     ReDim ary(0): ary(0) = value
End Sub

Sub papush(ary, ParamArray value()) 'this modifies parent ary object
    
    Dim i As Long, v
    
    If AryIsEmpty(ary) Then
        ReDim ary(0)
    Else
        push ary, ""
    End If
    
    i = UBound(ary)
    
    For Each v In value
        ary(i) = ary(i) & " " & v
    Next
    
End Sub
