#include <conio.h>
#include "vb.h"
#include <windows.h>
#include ".//pywin32_lib//PythonCOM.h"
#include <stdio.h>

//Author:  David Zimmer <dzzie@yahoo.com>
//Site:    http://sandsprite.com

//https://github.com/python/cpython/blob/main/Include/object.h
//https://py3c.readthedocs.io/en/latest/reference.html
//https://cffi.readthedocs.io/en/latest/embedding.html
//http://web.mit.edu/people/amliu/vrut/python/ext/ownershipRules.html
// 
//note: oleargs.cpp and ErrorUtils.cpp can manually import win32com..oleargs was changed to pycom, 
//      errorutils logger funcs are commented out..(redir to vbErr?)

//NOTE:  for the op_LoadFile to work this dll must compiled with /MD and use the same runtime version as the target python311.dll !
//       python311.dll uses VS2019 - VCRUNTIME140.dll, I do not like this design requirement but..
// 
//following does not work in vs2019 now...see define in pywrap.cpp, were just using def file now
#define EXPORT comment(linker, "/EXPORT:"__FUNCTION__"="__FUNCDNAME__)

#define PY_SSIZE_T_CLEAN
#include <Python.h>
#include <string>
#include <guiddef.h>
#include <comutil.h>
#include <algorithm>
#include <vector>

#pragma comment(lib, "comsuppw.lib")

#pragma warning(disable : 4996)  //depreceated function

//#define real_fwrite fwrite

vbCallback vbStdOut = 0;
vbHostResolverCallback vbHostResolver = 0;
vbDbgCallback vbLineInput = 0;
vbDbgCallback vbDbgOut = 0;
vbDbgCallback vbDbgWriteHandler = 0;

extern "C" PyObject * PyInit_pycomint(void);
extern  "C" PyObject * PyInit_pywinint(void);
extern void initPyWinTypes(bool init);
extern int PyCom_RegisterCoreSupport(void);
extern int PyCom_UnregisterCoreSupport(void);
extern void PyWinInterpreterState_Free();
extern void PyWinGlobals_Free();
extern BOOL PyCom_VariantFromPyObject(PyObject* obj, VARIANT* var); //PythonCOM.cpp
extern PyObject* PyCom_PyObjectFromVariant(const VARIANT* var);

char* mLastString = 0;
int isInit = 0;;
PyObject* mod = NULL;

enum opPy{
	op_Init = 0,
	op_Final = 1,
	op_LoadCode =2,
	op_GetFunc = 3,
	op_NewTuple = 4,
	op_CreateLong = 5,
	op_GetLong = 7,
	op_GetAttrStr = 12,
	op_AddIntConst = 13,
	op_ImportMod = 14,
	op_interrupt = 15,
	op_decref = 16,
	op_incref = 17,
	op_refcnt = 18,
	op_AddObj = 19,
	op_TypeName = 20,
	op_HasLen = 21,
	op_NewStr = 22,
	op_getMod = 23,
	op_eval = 24,
	op_LoadFile = 25
};

enum op_pyType {
	pt_String = 1,
	pt_Bool = 2,
	pt_List = 3,
	pt_Long = 4,
	pt_Tuple = 5,
	pt_Float = 6,
	pt_Dict = 7,
	pt_Complex = 8,
	pt_Bytes = 9,
	pt_ByteAry = 10,
	pt_None = 11,
	pt_isTrue = 12,
	pt_Set = 13
};

static PyObject* emb_print(PyObject* self, PyObject* args) {
	char* tmp;
	if (!PyArg_ParseTuple(args, "s", &tmp)) return NULL;
	if (vbStdOut) vbStdOut(cb_output, tmp);
	return Py_BuildValue("");
}

static PyObject* emb_err(PyObject* self, PyObject* args) {
	char* tmp;
	if (!PyArg_ParseTuple(args, "s",&tmp)) return NULL;
	if (vbStdOut) vbStdOut(cb_error, tmp);
	return Py_BuildValue("");
}

static PyObject* emb_refresh(PyObject* self, PyObject* args) {
	char* tmp;
	//if (!PyArg_ParseTuple(args, "s", &tmp)) return NULL;
	if (vbStdOut) vbStdOut(cb_Refresh,"");
	return Py_BuildValue("");
}

static PyObject* emb_dbgOut(PyObject* self, PyObject* args) {
	char* tmp; dcb_type i;
	if (!PyArg_ParseTuple(args, "is", &i, &tmp)) return NULL;
	if (vbDbgOut) vbDbgOut(i,tmp);
	return Py_BuildValue("");
}

static PyObject* emb_alert(PyObject* self, PyObject* args) {
	char* tmp;
	if (!PyArg_ParseTuple(args, "s", &tmp)) return NULL;
	MessageBoxA(0, tmp, "Python Script", 0);
	return Py_BuildValue("");
}


static PyMethodDef EmbMethods[] = {
	{"print", emb_print, METH_VARARGS, NULL},
	{"err", emb_err, METH_VARARGS, NULL},
	{"refresh", emb_refresh, METH_VARARGS, NULL},
	{"dbgOut", emb_dbgOut, METH_VARARGS, NULL},
	{"alert", emb_alert, METH_VARARGS, NULL},
	{NULL, NULL, 0, NULL}
};

static PyModuleDef EmbModule = {
	PyModuleDef_HEAD_INIT, "emb", NULL, -1, EmbMethods,
	NULL, NULL, NULL, NULL
};

static PyObject* PyInit_emb(void) {
	return PyModule_Create(&EmbModule);
}

int initPy() {

	if (isInit == 1) return -1;
	PyImport_AppendInittab("emb", &PyInit_emb); //add c functions to python
	
	//internal pythoncom startup
	int a = PyImport_AppendInittab("pycomint", &PyInit_pycomint); //should we just do this from the python package instead? we will need some pycode..
	int b = PyImport_AppendInittab("pywinint", &PyInit_pywinint); 
	initPyWinTypes(true);

	if (a == -1)MessageBoxA(0, "pycomint append failed!", "", 0);
	if (b == -1)MessageBoxA(0, "pywinint append failed!", "", 0);

	//Py_SetProgramName(argv[0]);
	// Py_SetPythonHome(python_home);
	Py_Initialize();

	//internal pythoncom startup
	PyWinGlobals_Ensure();
	PyCom_RegisterCoreSupport();

	mod = PyImport_ImportModule("__main__");
	if (mod == NULL) return -2;

	std::string stdOutErr = "\
import sys\n\
import emb\n\
sys.settrace(None) #remove if reloading\n\
class CatchOut:\n\
    def __init__(self):\n\
        self.value = ''\n\
    def write(self, txt):\n\
        emb.print(txt)\n\
    def flush(self):\n\
        pass\n\
class CatchErr:\n\
    def __init__(self):\n\
        self.value = ''\n\
    def write(self, txt):\n\
        emb.err(txt)\n\
    def flush(self):\n\
        pass\n\
def myExit(*args):\n\
    raise KeyboardInterrupt\n\
exit = myExit; quit = myExit; sys.exit = myExit;\n\
catchOut = CatchOut()\n\
catchErr = CatchOut()\n\
sys.stdout = catchOut\n\
sys.stderr = catchErr\n\
isDebugging = 0\n\
#def alert(x): #dont use embedded scripts here from user code, cause debugger jumps\n\
#    print('alert:'+str(x))\n\
def dumpGlobals():\n\
    d = globals(); tmp = '';\n\
    for k, v in d.items():\n\
        tmp = tmp + str(k) + '=' + str(v) + '<ITEM_DIVIDER>'\n\
    emb.dbgOut(2,tmp)\n\
def dumpLocals():\n\
    #print('in dumplocals')\n\
    try:\n\
        tmp = ''\n\
        if curFrame is None: return;\n\
        d = curFrame.f_locals\n\
        for k, v in d.items():\n\
            tmp = tmp + str(k) + '=' + str(v) + '<ITEM_DIVIDER>'\n\
        emb.dbgOut(3, tmp)\n\
    except Exception as e:\n\
         print('err in dumplocals:' + str(e))\n\
def dumpArgs():\n\
    #print('indumpArtgs')\n\
    try:\n\
        tmp = ''\n\
        if curFrame is None: return;\n\
        for i in range(curFrame.f_code.co_argcount):\n\
            n = curFrame.f_code.co_varnames[i]\n\
            v = curFrame.f_locals[n]\n\
            tmp = tmp + str(n) + '=' + str(v) + '<ITEM_DIVIDER>'\n\
        emb.dbgOut(4, tmp)\n\
    except Exception as e:\n\
         print('err in dumpArgs:' + str(e))\n\
"; //this is python code to redirect stdouts/stderr and hook exit/quit/sys.exit so they dont close the entire process!

/*
	def dumpLocals() :\n\
		print('in dumplocals')\n\
		try :\n\
		tmp = ''\n\
		if curFrame is None : return; \n\
			for i in range(curFrame.f_code.co_nlocals) :\n\
				n = curFrame.f_code.co_names[i]\n\
				v = curFrame.f_locals[n]\n\
				tmp = tmp + str(n) + '=' + str(v) + '<ITEM_DIVIDER>'\n\
				emb.dbgOut(3, tmp)\n\
				except Exception as e : \n\
				print('err in dumplocals:' + str(e))\n\*/

	PyRun_SimpleString(stdOutErr.c_str()); //invoke code to redirect

	isInit = 1;
	return 1;


}

//so ary() as string...is getting converted to ascii strings, the runtime is persistant..
int __stdcall SetArgv(VARIANT* v) {

	//https://learn.microsoft.com/en-us/windows/win32/api/wtypes/ne-wtypes-varenum
	//https://learn.microsoft.com/en-us/windows/win32/api/oaidl/ns-oaidl-safearray#:~:text=The%20fFeatures%20flags%20describe%20attributes%20of%20an%20array,freeing%20the%20array%20without%20referencing%20its%20containing%20variant.

	if (v->vt != (VT_BYREF | VT_ARRAY | VT_BSTR)) return -1; //0x6008 

	SAFEARRAY** psa;
	psa = v->pparray;

	/*
	char buf[1000];
	sprintf(buf, "%X", (*psa)->fFeatures);
	MessageBoxA(0, buf, "", 0);
	return -1;
	*/

	//accept only 1 dimensional 0 based string array
	if ((*psa)->cDims != 1) return -2;
	if ((*psa)->fFeatures != 0x180 /*FADF_BSTR | FADF_HAVEVARTYPE*/) return -3; 
	if ((*psa)->rgsabound[0].lLbound != 0) return -4;

	int lElements = (*psa)->rgsabound[0].cElements;
	if (lElements == 0) return -5;

	wchar_t** tmp;
	HRESULT hr = SafeArrayAccessData(*psa, (void**)&tmp);
	if (hr != S_OK) return -6;

	PySys_SetArgv(lElements, tmp);
	SafeArrayUnaccessData(*psa);

}


//if we declare as returning Variant in vb, first arg is a stack address..
int __stdcall GetStr(int pyObj, VARIANT* v) {

	VariantInit(v);
	
	if (pyObj == 0 || pyObj == -1) return 0;

	int sz = 0; 
	BSTR bstr;
	PyObject* s;
	PyObject* po = (PyObject*)(pyObj);

	//if its a string
	if (PyUnicode_Check(po) == 1) {
		sz = PyUnicode_GET_SIZE(po);
		bstr = SysAllocStringLen(PyUnicode_AS_UNICODE(po), (unsigned)sz);
		v->bstrVal = bstr;
		v->vt = VT_BSTR;
		return  1;
	}

	//if we can convert it to string with __repr__
	s = PyObject_Repr(po);
	if (s != NULL) {
		sz = PyUnicode_GET_SIZE(s);
		bstr = SysAllocStringLen(PyUnicode_AS_UNICODE(s), (unsigned)sz);
		v->bstrVal = bstr;
		v->vt = VT_BSTR;
		Py_DECREF(s);
		return  1;
	}

	//does it have a __str__ ?
	s = PyObject_Str(po);
	if(s != NULL){
		sz = PyUnicode_GET_SIZE(s);
		bstr = SysAllocStringLen(PyUnicode_AS_UNICODE(s), (unsigned)sz);
		v->bstrVal = bstr;
		v->vt = VT_BSTR;
		Py_DECREF(s);
		return  1;
	}

	/*PyObject* repr = PyObject_Repr(po);
	PyObject* str = PyUnicode_AsEncodedString(repr, "utf-8", "~E~");
	const char* bytes = PyBytes_AS_STRING(str);

	MessageBoxA(0, bytes, "", 0);

	Py_XDECREF(repr);
	Py_XDECREF(str);*/

	/*
	if (PyBytes_Check(po)) {

		PyObject* s = PyObject_Bytes(po);
		char* cs = PyBytes_AsString(s);
		sz = PyBytes_Size(s);
		//bstr = SysAllocStringLen(, (unsigned)sz);
		Py_XDECREF(s);
	}*/

	//PyObject* value = PyUnicode_FromEncodedObject(po, "utf-8", "~E~");

	//bytes, bytearray and other bytes-like objects are decoded according to the given encoding and using the error handling defined by errors. Both can be NULL to have the interface use the default values
	//PyObject* value = PyUnicode_FromEncodedObject(po, "utf-8", "~E~");
	//if (!value) return 0;

	
	//Py_DECREF(value);

	return  0;

}

//Mark saved my sanity :))
int __stdcall ObjToVariant(int pyObj, VARIANT* v) {
	if (pyObj == 0 || pyObj == -1) return -1;
	PyObject* po = (PyObject*)pyObj;
	return (int)PyCom_VariantFromPyObject(po, v);
}

int __stdcall VariantToPyObj(VARIANT *v) {
	return (int)PyCom_PyObjectFromVariant(v);
}

int __stdcall SetItem(PyObject* pArgs, int sz, PyObject* pVal) {
	return PyTuple_SetItem(pArgs, sz, pVal);
}

int __stdcall CallObject(PyObject* pFunc, PyObject* pArgs) {
	return (int)PyObject_CallObject(pFunc, pArgs);
}

//returns 0 for success, -1 for error
int __stdcall PyCheck(int operation, int arg1) {

	if (arg1 == -1 || arg1 == 0) return -1;
	PyObject* po = (PyObject*)arg1;

	switch (operation) {
		case pt_isTrue: return PyObject_IsTrue(po);
		case pt_None: return (po == Py_None ? 1 : 0);
		case pt_Long: return PyLong_Check(po);
		case pt_String: return  PyUnicode_Check(po);
		case pt_Bool: return Py_IS_TYPE(po, &PyBool_Type);
		case pt_List: return PyList_Check(po);
		case pt_Tuple: return PyTuple_Check(po);
		case pt_Float: return PyFloat_Check(po);
		case pt_Dict: return PyDict_Check(po);
		case pt_Complex: return PyComplex_Check(po);
		case pt_Bytes: return PyBytes_Check(po);
		case pt_ByteAry: return PyByteArray_Check(po);
		case pt_Set: return PySet_Check(po);
	}

	return -1;
}

//maybe its better to look for attr __len__ and call it, thats generic for any object that supports it, 
//this requires knowledge of types and intrinsic only..
int __stdcall PySize(int operation, int arg1) {

	if (arg1 == -1 || arg1 == 0) return -1;
	PyObject* po = (PyObject*)arg1;

	switch (operation) {
		case pt_String: return  PyUnicode_WSTR_LENGTH(po);
		case pt_List: return PyList_Size(po);
		case pt_Tuple: return PyTuple_Size(po);
		case pt_Dict: return PyDict_Size(po);
		case pt_Bytes: return PyBytes_Size(po);
		case pt_ByteAry: return PyByteArray_Size(po);
		case pt_Set: return PySet_Size(po);
	}

	return -2;
}

bool FileExists(char* szPath)
{
	DWORD dwAttrib = GetFileAttributesA(szPath);
	bool rv = (dwAttrib != INVALID_FILE_ATTRIBUTES && !(dwAttrib & FILE_ATTRIBUTE_DIRECTORY)) ? true : false;
	return rv;
}

std::string FileNameFromPath(char* path) {

	if (path == NULL || strlen(path) == 0) return 0;

	std::string ret;
	std::string s = path;

	int i, sz;
	int x = s.find("/");
	if (x >= 0) {
		std::replace(s.begin(), s.end(), '/', '\\');
	}

	x = s.find_last_of("\\");
	if (x >= 0) {
		x++;
	}

	sz = s.length() - x;
	if (sz < 1) return 0;

	for (i = 0; i < sz; i++) {
		if (path[x + i] == '?') break;
		ret += path[x + i];
	}
	return ret;
}


//generally returns 0 for success, -1 for error
int __stdcall PyOp(int operation, int arg1, char* arg2, int arg3){
#pragma EXPORT
	
	int rv = 0;
	PyObject* po = NULL;

	/*these do not require a context..
	switch(operation){
		case opd_LastString: return mLastString;
		case opd_ScriptTimeout: 
			watchdogTimeout = arg1; return 0;
	}*/

	
	switch (operation) {
		case op_Init: return initPy();
		case op_LoadCode: return PyRun_SimpleString(arg2); // 0 on success or -1 if an exception was raised.
		case op_interrupt: PyErr_SetInterrupt(); return 0;//Py_Exit(arg1); 
		case op_NewStr: return (int)PyUnicode_FromString(arg2);
		case op_decref: if (arg1 == 0 || arg1 == -1) return -1; else { Py_XDECREF((PyObject*)arg1); return ((PyObject*)arg1)->ob_refcnt; }
		case op_incref: if (arg1 == 0 || arg1 == -1) return -1; else { Py_XINCREF((PyObject*)arg1); return ((PyObject*)arg1)->ob_refcnt; }
		case op_refcnt: if (arg1 == 0 || arg1 == -1) return -1; else return ((PyObject*)arg1)->ob_refcnt;

		case op_LoadFile: //Note: python311.dll and this dll must both be compiled with /MD to share runtime FILE* lookups
		{                 //      also required they share same runtime version (VS2019 VCRUNTIME140.dll) kind of subtleand delicate :(
			if (!FileExists(arg2)) return 0;  
			FILE* fp = fopen(arg2, "rb");
			if (fp == NULL) return 0;
			//std::string fname = FileNameFromPath(arg2);
			return PyRun_SimpleFileExFlags(fp, arg2, 1, NULL); //file handle closed by function
		}

		case op_eval:
		{
			PyObject* code = Py_CompileString(arg2, "eval", Py_eval_input);
			if ((int)code == 0) return 0;
			PyObject* global_dict = PyModule_GetDict((arg1 == 0 ? mod : (PyObject*)arg1));
			PyObject* local_dict = PyDict_New();
			PyObject* obj = PyEval_EvalCode(code, global_dict, local_dict);
			Py_XDECREF(code);
			Py_XDECREF(local_dict);
			return (int)obj;
		}

		case op_getMod: 
			po = PyUnicode_FromString(arg2);
			if ((int)po == 0) return 0;
			rv = (int)PyImport_GetModule(po);
			Py_XDECREF(po);
			return rv;

		case op_Final: 
			mod = NULL; 
			isInit = 0; 
			PyCom_UnregisterCoreSupport();
			initPyWinTypes(false);
			PyWinInterpreterState_Free();
			PyWinGlobals_Free();

			return Py_FinalizeEx();

		case op_TypeName:
			if (arg1 == -1 || arg1 == 0) return -1;
			return (int)((PyObject*)arg1)->ob_type->tp_name; //this will end poorly if you pass junk arg

		case op_HasLen:
			if (arg1 == -1 || arg1 == 0) return -1;
			return (int)((PyObject*)arg1)->ob_type->tp_as_sequence; //this shows if object suports __len__

		case op_ImportMod:
			if (mod == NULL) return -1;
			po = PyImport_ImportModule(arg2); //PyImport_AddModule(arg2);
			PyModule_AddObject(mod, arg2, po);
			Py_INCREF(po);
			return (int)po;

		case op_AddObj: //we could use another int arg to specify a different obj other than main module..this is ok for global obj constants though.
			if (mod == NULL) return -1;
			return PyModule_AddObject(mod, arg2, (PyObject*)arg1);

		case op_AddIntConst:
			if (mod == NULL) return -1;
			return PyModule_AddIntConstant(mod, arg2, arg1);

		case op_GetAttrStr:
			po = PyObject_GetAttrString((arg1==0 ? mod : (PyObject*)arg1), arg2); //does not need decref?
			return (int)po;

		case op_GetLong: 
			if (arg1 == -1 || arg1 == 0) return -1;
			po = (PyObject*)arg1;
			return PyLong_AsLong(po);

		case op_GetFunc: 
				po = PyObject_GetAttrString(mod, arg2); 
				if (po && PyCallable_Check(po)) return (int)po;
				return 0;

		case op_NewTuple:
				po = PyTuple_New(arg1);
				return (int)po;

		case op_CreateLong:
				po = PyLong_FromLong(arg1);
				return (int)po;

	}

	return -1;

}

void __stdcall SetCallBacks(void* lpfnMsgHandler, void* lpfnDbgHandler, void* lpfnHostResolver, void* lpfnLineInput, void* lpfnDbgWriteHandler){
#pragma EXPORT
	vbStdOut     = (vbCallback)lpfnMsgHandler;
	vbDbgOut = (vbDbgCallback)lpfnDbgHandler;
	vbHostResolver = (vbHostResolverCallback)lpfnHostResolver;
	vbLineInput = (vbDbgCallback)lpfnLineInput;
	vbDbgWriteHandler = (vbDbgCallback)lpfnDbgWriteHandler;
}

int __stdcall setLastString(const char* s){ //accepts null string to just free last buffer
#pragma EXPORT
	if(mLastString != 0){ free(mLastString); mLastString = 0 ;}
	if(s==0) return 0;
	mLastString = strdup(s);
	return strlen(s);
}

 
VARIANT __stdcall About_py4vb() {

	char* s = "py4vb is a helper dll to access python api as stdcall exports.\n\n" \
		"this dll also includes the COM integration code from Mark Hammonds " \
		"pywin32 project for ease of access. See License.txt in the pywin32 dir\n\n" \
		"This project is being developed by David Zimmer and hosted on Sandsprite.com\n\n" \
		"Please see licenses from all respective authors.";

	VARIANT v;
	VariantInit(&v);

	BSTR ret = _com_util::ConvertStringToBSTR(s);
	v.vt = VT_BSTR;
	v.bstrVal = ret;
	return v;

}



/*
int comResolver(duk_context *ctx) {
	int i, hasRetVal ;
	const char* meth = 0;
	int realArgCount = 0;
	int hInst = 0;

	int n = duk_get_top(ctx);  //padded number of args..not usable for us here..
	
	if(n < 3) return 0; //we require at least 3 args for this function..
	if(vbHostResolver==NULL) return 0; 
	
	meth = duk_safe_to_string(ctx, 0);   //arg0 is obj.method string
	realArgCount = duk_to_number(ctx,1); //arg1 is arguments.length
	hInst = duk_to_number(ctx,2);       //arg2 is this com objects hinst variable if not a top level obj (0 if not)
	hasRetVal = vbHostResolver(meth, ctx, realArgCount, hInst);

	if(hasRetVal != 0 && hasRetVal != 1){
			MessageBox(0,"comresolver","vbdev the hasRetVal must be 0 or 1",0);
			hasRetVal = 1;
	}

	return hasRetVal;  
}

 */
